
/* extract.c - utility to remove silence from beginning and end of samples */
/* written by Istvan Varga, May 11 2002 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define LIMIT1	256	/* limit at beginning of file */
#define ENVLEN	512UL	/* release envelope length */

/* sound file structure */

typedef struct {
	unsigned long	sr;		/* sample rate */
	unsigned long	nsmps;		/* length in samples */
	short		*data;		/* sample data */
} SNDFL;

/* read / write little-endian integers */

unsigned long	read_int_le (FILE *f, int bytes)
{
	unsigned long	n;
	int		i;

	if (!bytes) return 0UL;
	n = (unsigned long) getc (f);
	i = 8;
	while (--bytes) {
		n += (unsigned long) getc (f) << i;
		i += 8;
	}
	return n;
}

void	write_int_le (FILE *f, unsigned long n, int bytes)
{
	while (bytes--) {
		putc ((int) (n & 0xFFUL), f);
		n >>= 8;
	}
}

/* open WAV file and load into memory */

int	open_wav_file (FILE *f, SNDFL *s)
{
	unsigned long	n;

	/* check header */
	n = read_int_le (f, 4);
	if (n != 0x46464952UL) return -1;		/* "RIFF" */
	n = read_int_le (f, 4);
	n = read_int_le (f, 4);
	if (n != 0x45564157UL) return -1;		/* "WAVE" */
	n = read_int_le (f, 4);
	if (n != 0x20746D66UL) return -1;		/* "fmt " */
	n = read_int_le (f, 4);
	if (n != 0x00000010UL) return -1;		/* format length */
	n = read_int_le (f, 2);
	if (n != 0x00000001UL) return -1;		/* wave format */
	n = read_int_le (f, 2);
	if (n != 0x00000001UL) return -1;		/* nr of channels */
	n = read_int_le (f, 4);
	if (n < 1000UL || n > 1000000UL) return -1;	/* sample rate */
	s->sr = n;
	n = read_int_le (f, 4);
	n = read_int_le (f, 2);
	if (n != 0x00000002UL) return -1;		/* block alignment */
	n = read_int_le (f, 2);
	if (n != 0x00000010UL) return -1;		/* bits per sample */
	n = read_int_le (f, 4);
	if (n != 0x61746164UL) return -1;		/* "data" */
	n = read_int_le (f, 4);
	if (n < (ENVLEN << 1)) return -1;		/* length */
	s->nsmps = n >> 1;
	/* allocate memory for sample data */
	s->data = (short *) malloc (s->nsmps << 1);
	if (s->data == NULL) return -1;
	/* load file into memory */
	for (n = 0UL; n < s->nsmps; n++) s->data[n] = read_int_le (f, 2);

	return 0;
}

/* write WAV file */

void	write_wav_file (FILE *f, SNDFL *s)
{
	unsigned long	n;

	write_int_le (f, 0x46464952UL, 4);		/* "RIFF" */
	n = (s->nsmps << 1) + 36UL;
	write_int_le (f, n, 4);
	write_int_le (f, 0x45564157UL, 4);		/* "WAVE" */
	write_int_le (f, 0x20746D66UL, 4);		/* "fmt " */
	write_int_le (f, 0x00000010UL, 4);		/* format length */
	write_int_le (f, 0x00000001UL, 2);		/* wave format */
	write_int_le (f, 0x00000001UL, 2);		/* nr of channels */
	n = s->sr;
	write_int_le (f, n, 4);				/* sample rate */
	n <<= 1;
	write_int_le (f, n, 4);				/* bytes per second */
	write_int_le (f, 0x00000002UL, 2);		/* block alignment */
	write_int_le (f, 0x00000010UL, 2);		/* bits per sample */
	write_int_le (f, 0x61746164UL, 4);		/* "data" */
	n = s->nsmps << 1;
	write_int_le (f, n, 4);				/* length */
	/* write data to file */
	for (n = 0UL; n < s->nsmps; n++) write_int_le (f, s->data[n], 2);
}

/* close WAV file */

void	close_wav_file (SNDFL *s)
{
	s->sr = s->nsmps = 0UL;
	free (s->data);
	s->data = NULL;
}

int	main (int argc, char **argv)
{
	FILE		*infile, *outfile;
	SNDFL		s;
	unsigned long	i, strt, end;
	long		x;
	short		limit2;

	if (argc != 5) {
		fprintf (stderr, "extract: ");
		fprintf (stderr, "insufficient or too many arguments\n");
		fprintf (stderr, "usage: extract <infile> <outfile> <sr> ");
		fprintf (stderr, "<end_limit>\n");
		exit (-1);
	}
	/* open files */
	infile = fopen (argv[1], "rb");
	if (infile == NULL) {
		fprintf (stderr, "extract: cannot open input file\n");
		exit (-1);
	}
	outfile = fopen (argv[2], "wb");
	if (outfile == NULL) {
		fprintf (stderr, "extract: cannot open output file\n");
		exit (-1);
	}
	/* limit at end of file */
	limit2 = (short) abs ((int) atoi (argv[4]));
	/* read input file */
	if (open_wav_file (infile, &s) < 0) {
		fprintf (stderr, "extract: error reading input file\n");
		exit (-1);
	}
	/* start time */
	i = 0UL;
	while (i < (s.nsmps - ENVLEN) && abs (s.data[i]) < LIMIT1) i++;
	if (i) i--;
	strt = i;
	/* end time */
	i = s.nsmps;
	do {
		if (abs (s.data[--i]) >= limit2) break;
	} while (i > strt);
	i += ENVLEN;
	end = (i < s.nsmps ? i : s.nsmps);
	/* extract */
	if (strt) {
		for (i = strt; i < end; i++) s.data[i - strt] = s.data[i];
	}
	s.nsmps = end - strt;
	/* release envelope */
	for (i = 1; i < ENVLEN; i++) {
		x = (long) s.data[s.nsmps - i];
		x *= (long) i;
		x += (x < 0L ? -((long) (ENVLEN >> 1)) : (long) (ENVLEN >> 1));
		x /= (long) ENVLEN;
		s.data[s.nsmps - i] = (short) x;
	}
	/* set new sample rate if specified on command line */
	i = (unsigned long) atoi (argv[3]);
	if (i) {
		if (i < 1000UL || i > 1000000UL) {
			fprintf (stderr, "extract: invalid sample rate\n");
			exit (-1);
		}
		s.sr = i;
	}
	/* write output file */
	write_wav_file (outfile, &s);
	/* close files */
	close_wav_file (&s);
	fclose (infile);
	fflush (outfile); fclose (outfile);

	return 0;
}

